# Do NOT edit this file.
# Date: April 12, 2013
# SRBX Version: 1

#******************************************************************************

.Last <- function()
{
    stop( sprintf( .SASGetMessage("Stop_CantExitR.txt"), .Platform$GUI ) )
}

#******************************************************************************

.SASImportDataExchangeFile <- function( PathName )
{
    PrevWarn <- getOption( "warn" )
    options( warn=-1 )
    on.exit( options( warn=PrevWarn ), add=TRUE )

    fc <- file( description=PathName, open="rb", blocking=TRUE, encoding="" )
    on.exit( close( fc ), add=TRUE )

    EncNames <- c( "unknown", "latin1", "UTF-8" )



    # default.stringsAsFactors() removed from R 4.3.0+
    r_version = R.Version()
    r_version_major = as.numeric(r_version$major)
    r_version_minor = as.numeric(r_version$minor)

    if ( (r_version_major < 4) | (r_version_major == 4 & r_version_minor < 3.0)) {
        MakeFactors <- default.stringsAsFactors()
    } else {
        MakeFactors <- FALSE
    }

    FileHeader <- readBin( fc, "integer", n=8 )

    if ( FileHeader[1] != 1 )
        stop( .SASGetMessage("Stop_UnsupportedDataExchangeFileFormat.txt") )

    NumCols <- FileHeader[2]
    NumRows <- FileHeader[3]

    vecs <- vector( "list", NumCols )

    ColNames <- vector( "character", NumCols )

    i <- 1
    while ( i <= NumCols ) {
        ColHeader <- readBin( fc, "integer", n=8 )

        ColName <- readBin( fc, "character", n=1 )
        if ( ColHeader[3] > 1 )
            Encoding(ColName) <- EncNames[ColHeader[3]]
        ColNames[i] <- ColName

        if ( ColHeader[1] == 0 ) {
            x <- readBin( fc, "double", n=NumRows )
            if ( ColHeader[2] == 1 )
                x <- factor( x )
            }
        else if ( ColHeader[1] == 1 ) {
            x <- readBin( fc, "integer", n=NumRows )
            if ( ColHeader[2] == 1 )
                x <- factor( x )
            }
        else if ( ColHeader[1] == 2 ) {
            x <- readBin( fc, "logical", n=NumRows )
            }
        else if ( ColHeader[1] == 3 ) {
            x <- readBin( fc, "character", n=NumRows )
            x[which(x=="\x1F")] <- NA_character_
            if ( ColHeader[4] > 1 )
                Encoding(x) <- EncNames[ColHeader[4]]
            if ( MakeFactors )
                x <- factor( x )
            }
        else if ( ColHeader[1] == 4 ) {
            x <- readBin( fc, "double", n=NumRows )
            class(x) <- "Date"
            if ( ColHeader[2] == 1 )
                x <- factor( x )
            }
        else if ( ColHeader[1] == 5 ) {
            x <- readBin( fc, "double", n=NumRows )
            if ( .SASRVersion >= 21100 )
                x <- .POSIXct( x, tz="UTC" )
            else {
                class(x) <- c( "POSIXt", "POSIXct" )
                attr( x, "tzone" ) <- "UTC"
                }
            if ( ColHeader[2] == 1 )
                x <- factor( x )
            }
        else if ( ColHeader[1] == 6 ) {
            x <- readBin( fc, "integer", n=NumRows )
            class(x) <- "Date"
            if ( ColHeader[2] == 1 )
                x <- factor( x )
            }
        else if ( ColHeader[1] == 7 ) {
            x <- readBin( fc, "integer", n=NumRows )
            if ( .SASRVersion >= 21100 )
                x <- .POSIXct( x, tz="UTC" )
            else {
                class(x) <- c( "POSIXt", "POSIXct" )
                attr( x, "tzone" ) <- "UTC"
                }
            if ( ColHeader[2] == 1 )
                x <- factor( x )
            }
        else
            stop( .SASGetMessage("Stop_UnsupportedDataExchangeFileFormat.txt") )

        vecs[[i]] <- x

        i <- i + 1
        }

    ColNames <- make.names( ColNames, TRUE )

    return (structure( vecs, names=ColNames, row.names=c( NA, -NumRows ), class="data.frame" ))
}

#******************************************************************************

.SASExportDataExchangeFile <- function( x, PathName )
{
    ExpandEmbeddedMatrices <- function( df )
    {
        VarNames <- names( df )

        NumVar <- dim(df)[2]
        NumRow <- dim(df)[1]

        vecs <- vector( "list", NumVar )

        VecNames <- vector( "character", NumVar )

        i <- 1
        k <- 1
        while ( i <= NumVar ) {
            x <- df[[i]]

            if ( length( dim(x) ) < 2 ) {
                vecs[[k]] <- x
                VecNames[k] <- VarNames[i]
                k <- k + 1
                }
            else if ( is.matrix(x) ) {
                NumCol <- dim(x)[2]
                ColNames <- dimnames( x )[[2]]
                if ( is.null( ColNames ) )
                    ColNames <- as.character( 1:NumCol )
                else
                    ColNames[which( is.na( ColNames ) )] <- "_"
                c <- 1
                while ( c <= NumCol ) {
                    vecs[[k]] <- x[,c]

                    if ( NumCol == 1 )
                        VecNames[k] <- VarNames[i]
                    else
                        VecNames[k] <- paste( VarNames[i], ".", ColNames[c], sep="" )

                    k <- k + 1
                    c <- c + 1
                    }
                }
            else
                stop( sprintf( .SASGetMessage("Stop_CantImportColumnContainsUnsupportedObject.txt"), VarNames[i] ) )

            i <- i + 1
            }

        return (structure( vecs, names=VecNames, row.names=c( NA, -NumRow ), class="data.frame" ))
    }

    #--------------------------------------------------------------------------

    AddVarsForTimeSeries <- function( df )
    {
        VarNames <- names( df )

        OriginalNumVar <- dim(df)[2]

        LastVarIndex <- OriginalNumVar
        i <- 1
        while ( i <= OriginalNumVar ) {
            v <- df[[i]]

            if ( is.ts(v) ) {
                LastVarIndex <- LastVarIndex + 1
                df[LastVarIndex] <- as.vector( time( v ) )
                name <- paste( VarNames[i], "_ts", sep="" )
                names(df)[LastVarIndex] <- name
                }

            i <- i + 1
            }

        return (df)
    }

    #--------------------------------------------------------------------------

    WriteSRBXFile <- function( df, PathName )
    {
        PrevWarn <- getOption( "warn" )
        options( warn=-1 )
        on.exit( options( warn=PrevWarn ), add=TRUE )

        fc <- file( description=PathName, open="wb", blocking=TRUE, encoding="" )
        on.exit( close( fc ), add=TRUE )

        ColNames <- names( df )

        NumCols <- dim(df)[2]

        FileHeader <- vector( "double", 8 )
        FileHeader[1] <- 1
        FileHeader[2] <- NumCols
        FileHeader[3] <- dim(df)[1]

        writeBin( as.integer(FileHeader), fc )

        ColHeader <- vector( "double", 8 )

        i <- 1
        while ( i <= NumCols ) {
            ColName <- ColNames[i]
            ColHeader[3] <- 1

            x <- df[[i]]

            if ( is.factor(x) ) {
                ColHeader[2] <- 1
                xlevels <- levels( x )
                xlevels_num <- as.double( xlevels )
                if ( all( is.na( xlevels ) == is.na( xlevels_num ) ) ) {
                    x <- xlevels_num[x]
                    ColHeader[1] <- 0
                    ColHeader[4] <- 0
                    }
                else {
                    x <- xlevels[x]
                    x[is.na(x)] <- "\x1F"
                    ColHeader[1] <- 3
                    ColHeader[4] <- 1
                    }
                }
            else {
                if ( is.double(x) ) {
                    ColHeader[2] <- 0
                    ColHeader[4] <- 0
                    if ( is( x, "Date" ) )
                        ColHeader[1] <- 4
                    else if ( is( x, "POSIXt" ) )
                        ColHeader[1] <- 5
                    else
                        ColHeader[1] <- 0
                    }
                else if ( is.integer(x) ) {
                    ColHeader[2] <- 0
                    ColHeader[4] <- 0
                    if ( is( x, "Date" ) )
                        ColHeader[1] <- 6
                    else if ( is( x, "POSIXt" ) )
                        ColHeader[1] <- 7
                    else
                        ColHeader[1] <- 1
                    }
                else if ( is.logical(x) ) {
                    ColHeader[1] <- 2
                    ColHeader[2] <- 1
                    ColHeader[4] <- 0
                    }
                else if ( is.character(x) ) {
                    x[is.na(x)] <- "\x1F"
                    ColHeader[1] <- 3
                    ColHeader[2] <- 1
                    ColHeader[4] <- 1
                    }
                else if ( is.complex(x) )
                    stop( sprintf( .SASGetMessage("Stop_CantImportColumnContainsComplexNumbers.txt"), ColName ) )
                else if ( is.raw(x) )
                    stop( sprintf( .SASGetMessage("Stop_CantImportColumnContainsRawData.txt"), ColName ) )
                else
                    stop( sprintf( .SASGetMessage("Stop_CantImportColumnContainsUnsupportedObject.txt"), ColName ) )
                }

            writeBin( as.integer(ColHeader), fc )
            writeBin( ColName, fc )
            attributes(x) <- NULL
            writeBin( x, fc )

            i <- i + 1
            }
    }

    #--------------------------------------------------------------------------

    if ( is( try( q <- parse( text=x ), silent=TRUE ), "try-error" ) )
        stop( .SASGetMessage("Stop_InvalidRExpr.txt") )

    x____ <- eval.parent( q, 1 )

    df <- as.data.frame( x____ )

    NumCol <- dim(df)[2]

    if ( NumCol == 0 )
        stop( .SASGetMessage("Stop_RExprEvalNull.txt") )

    if ( is.null( names(df) ) ) {
        if ( NumCol == 1 )
            names(df) <- "X"
        else
            names(df) <- paste( "X", 1:NumCol, sep="" )
        }
    else {
        q <- is.na( names(df) )
        if ( any( q ) )
            names(df)[which( q )] <- "X_"
        if ( names(df)[1] == "x____" )
            names(df)[1] <- x
        }

    if ( any( sapply( df, function( a ) length( dim(a) ) == 2 && dim(a)[2] > 1 ) ) )
        df <- ExpandEmbeddedMatrices( df )

    if ( any( sapply( df, function( a ) is.ts(a) ) ) )
        df <- AddVarsForTimeSeries( df )

    #--------------------------------------------------------------------------

    WriteSRBXFile( df, PathName )
}

#******************************************************************************

.SASGetMessage <- function( key )
{
    msg <- .SASMessages$Values[which( .SASMessages$Keys == key )]

    if ( length(msg) == 0 )
        msg <- "The localized message is missing!"

    if ( Encoding(msg) == "UTF-8" )
        msg <- iconv( msg, "UTF-8", "", "?" )

    return (msg)
}

#******************************************************************************

do.call( function(x)
{
    LoadPropertiesFile <- function( PathName )
    {
        df <- read.table( PathName, quote=NULL, sep="\x1F", fill=TRUE, comment.char="", colClasses="character", na.strings="\x1F", fileEncoding="UTF-8" )
        cv <- df[[1]]

        # Filter out non-assignment lines
        cv <- grep( "^[ \t]*[A-Za-z0-9_.]+[ \t]*=.*", cv, value=TRUE )

        # Remove leading whitespace
        cv <- sub( "^[ \t]*", "", cv )

        Keys   <- sub( "[ \t]*=.*",                    "", cv )
        Values <- sub( "^[A-Za-z0-9_.]+[ \t]*=[ \t]*", "", cv )

        return (data.frame( Keys, Values, stringsAsFactors=FALSE ))
    }

    #--------------------------------------------------------------------------

    GetPropertyString <- function( df, key )
    {
        return (df$Values[which( df$Keys == key )])
    }

    #--------------------------------------------------------------------------

    GetLocaleSuffix_POSIX <- function()
    {
        locale <- Sys.getlocale( "LC_CTYPE" )

        locale <- sub( "[.].*", "", locale )

        suffix <- sub( "_.*", "", locale )

        if ( suffix == "zh" ) {
            df <- LoadPropertiesFile( "tkIntR.properties" )
            suffix <- GetPropertyString( df, locale )
            }

        if ( suffix == "C" )
            suffix <- "en"

        return (suffix)
    }

    #--------------------------------------------------------------------------

    GetLocaleSuffix_Windows <- function()
    {
        cp <- l10n_info()$codepage

        if ( cp == 936 )
            key <- "ChineseSimplified"
        else if ( cp == 950 )
            key <- "ChineseTraditional"
        else
            key <- sub( "[_ (-].*", "", Sys.getlocale( "LC_CTYPE" ) )

        df <- LoadPropertiesFile( "tkIntR.properties" )

        return (GetPropertyString( df, key ))
    }

    #--------------------------------------------------------------------------

    if ( .Platform$OS.type == "windows" )
        suffix <- GetLocaleSuffix_Windows()
    else
        suffix <- GetLocaleSuffix_POSIX()

    if ( length(suffix) == 0 )
        filename <- "tkIntR_en.properties"
    else {
        filename <- paste( "tkIntR_", suffix, ".properties", sep="" )

        if ( file.access( filename, 0 ) != 0 )
            filename <- "tkIntR_en.properties"
        }

    df <- LoadPropertiesFile( filename )

    assign( ".SASMessages", df, envir=.GlobalEnv )

    ver <- unclass( getRversion() )[[1]]
    ver <- 10000 * ver[1] + 100 * ver[2] + ver[3]
    if ( ver < 21300 )
        stop( sprintf( .SASGetMessage("Stop_UnsupportedRVersion.txt"), 2L, 13L, 0L ) )
    assign( ".SASRVersion", ver, envir=.GlobalEnv )

}, list(0) )
